from bobocep.forwarder.bobo_forwarder import \
    BoboForwarder
from bobocep.rules.actions.bobo_action import BoboAction
from bobocep.rules.events.composite_event import CompositeEvent


class ActionForwarder(BoboForwarder):
    """An event forwarder that performs an action with a new CompositeEvent
    instance before passing it to the subscribers of the forwarder.

    :param action: The action to perform.
    :type action: BoboAction

    :param max_queue_size: The maximum data queue size,
                           defaults to 0 (infinite).
    :type max_queue_size: int, optional

    :param active: Whether task should start in an active state,
                   defaults to True.
    :type active: bool, optional
    """

    def __init__(self,
                 action: BoboAction,
                 max_queue_size: int = 0,
                 active: bool = True) -> None:
        super().__init__(max_queue_size=max_queue_size, active=active)

        self._action = action

    def _handle_forwarder_event(self, event: CompositeEvent) -> bool:
        return self._action.execute(event).success



from abc import ABC, abstractmethod


class BoboNullData(ABC):
    """
    An abstract class to define periodic null data.
    """

    def __init__(self) -> None:
        super().__init__()

    @abstractmethod
    def get_null_data(self):
        """
        :return: The null data.
        """



from typing import Dict, Tuple

from bobocep.rules.events.bobo_event import BoboEvent


class MatchEvent:
    """
    A BoboEvent instance that has been selected as a match for some state
    criteria by one or more runs.

    :param nfa_name: The name of the BoboNFA instance.
    :type nfa_name: str

    :param label: The state label with which the event is associated.
    :type label: str

    :param event: The event selected as being a match for some state.
    :type event: BoboEvent

    :param next_ids: Matches from older states, to which this event links,
                     defaults to an empty dict.
    :type next_ids: Dict[str, str], optional

    :param previous_ids: Matches from newer states, to which this event
                         links, default to an empty dict.
    :type previous_ids: Dict[str, str], optional
    """

    NFA_NAME = "nfa_name"
    LABEL = "label"
    EVENT = "event"
    NEXT_IDS = "next_ids"
    PREVIOUS_IDS = "previous_ids"

    def __init__(self,
                 nfa_name: str,
                 label: str,
                 event: BoboEvent,
                 next_ids: Dict[str, Tuple[str, str]] = None,
                 previous_ids: Dict[str, Tuple[str, str]] = None) -> None:
        super().__init__()

        self.nfa_name = nfa_name
        self.label = label
        self.event = event
        self.next_ids = {} if next_ids is None else next_ids
        self.previous_ids = {} if previous_ids is None else previous_ids

    def add_pointer_next(self,
                         version: str,
                         label: str = "",
                         event_id: str = "") -> None:
        """
        Points match event to the next match event ID using a run version.

        :param version: The run version.
        :type version: str

        :param label: The state label, defaults to an empty string.
        :type label: str, optional

        :param event_id: The event ID to point to, defaults to an empty string.
        :type event_id: str, optional

        :raises RuntimeError: Label and event ID are the same as those in the
                              match event i.e. attempt to point match event to
                              itself.
        """

        if label == self.label and event_id == self.event.event_id:
            raise RuntimeError("Cannot point match event to itself.")

        self.next_ids[version] = (label, event_id)

    def add_pointer_previous(self,
                             version: str,
                             label: str = "",
                             event_id: str = "") -> None:
        """
        Points match event to the previous match event ID using a run version.

        :param version: The run version.
        :type version: str

        :param label: The state label, defaults to an empty string.
        :type label: str, optional

        :param event_id: The event ID to point to, defaults to an empty string.
        :type event_id: str, optional

        :raises RuntimeError: Label and event ID are the same as those in the
                              match event i.e. attempt to point match event to
                              itself.
        """

        if label == self.label and event_id == self.event.event_id:
            raise RuntimeError("Cannot point match event to itself.")

        self.previous_ids[version] = (label, event_id)

    def remove_all_pointers(self, version: str) -> None:
        """
        Removes all pointers to a match event with a given run version.

        :param version: The run version.
        :type version: str
        """

        self.next_ids.pop(version, None)
        self.previous_ids.pop(version, None)

    def has_pointers(self) -> bool:
        """
        :return: True if the match event points to any other match events,
                 False otherwise.
        """

        return len(self.next_ids) > 0 or len(self.previous_ids) > 0

    def to_dict(self) -> dict:
        """
        :return: A dict representation of the object.
        """

        return {
            self.NFA_NAME: self.nfa_name,
            self.LABEL: self.label,
            self.EVENT: self.event.to_dict(),
            self.NEXT_IDS: self.next_ids,
            self.PREVIOUS_IDS: self.previous_ids
        }



from bobocep.rules.actions.bobo_action import BoboAction
from bobocep.rules.nfas.patterns.bobo_pattern import BoboPattern


class BoboComplexEvent:
    """A complex event definition. It contains the name of a complex event,
    the pattern that, if identified, infers the existence of the complex event,
    and the actions to take if the complex event were to be identified.

    :param name: The complex event name.
    :type name: str

    :param pattern: The complex event pattern.
    :type pattern: BoboPattern

    :param action: The action to perform, defaults to no action.
    :type action: BoboAction, optional
    """

    def __init__(self,
                 name: str,
                 pattern: BoboPattern,
                 action: BoboAction = None) -> None:
        super().__init__()

        self.name = name
        self.pattern = pattern
        self.action = action



from bobocep.receiver.generators.data.bobo_null_data import BoboNullData


class BoboNullDataNone(BoboNullData):
    """
    Null data class that always returns None.
    """

    def __init__(self) -> None:
        super().__init__()

    def get_null_data(self):
        return None



from typing import Dict

from bobocep.rules.events.bobo_event import BoboEvent
from bobocep.rules.events.histories.bobo_history import BoboHistory


class CompositeEvent(BoboEvent):
    """A composite event.

    :param timestamp: The event timestamp indicating when it was first
                      generated.
    :type timestamp: int

    :param name: The event name.
    :type name: str

    :param history: The history of events that caused the composite event to be
                    generated.
    :type history: BoboHistory

    :param data: The event data, defaults to an empty dict.
    :type data: Dict[str, str], optional

    :param event_id: The event ID, defaults to a randomly generated ID.
    :type event_id: str, optional
    """

    NAME = "name"
    HISTORY = "history"

    def __init__(self,
                 timestamp: int,
                 name: str,
                 history: BoboHistory,
                 data: Dict[str, str] = None,
                 event_id: str = None) -> None:
        super().__init__(timestamp=timestamp,
                         data=data,
                         event_id=event_id)

        self.name = name
        self.history = history

    def to_dict(self) -> dict:
        """
        :return: A dict representation of the object.
        """

        return {
            self.TIMESTAMP: self.timestamp,
            self.NAME: self.name,
            self.HISTORY: self.history.to_dict(),
            self.DATA: self.data,
            self.EVENT_ID: self.event_id
        }



from threading import Thread, Event

from bobocep.setup.task.bobo_task import BoboTask


class BoboTaskThread(Thread):
    """
    A thread to run BoboTask instances. It runs the setup task once, and
    then runs the loop task continually until the thread is cancelled.
    There is a delay between each loop.

    :param task: The task to perform.
    :type task: BoboTask

    :param delay: The delay between loops.
    :type delay: float
    """

    def __init__(self,
                 task: BoboTask,
                 delay: float):
        Thread.__init__(self)

        self._task = task
        self._delay = delay
        self._cancel_event = Event()

    def run(self):
        """Runs the thread."""

        try:
            if (not self._cancel_event.is_set()) and \
                    (not self._task.is_cancelled()):
                self._task.setup()

                # thread has been cancelled if True
                while not self._cancel_event.wait(self._delay):
                    # task has been cancelled if True
                    if not self._task.is_cancelled():
                        self._task.loop()
                    else:
                        break
        except Exception:
            pass
        finally:
            try:
                self._task.cancel()
            except Exception:
                pass
            self._cancel_event.set()

    def is_cancelled(self) -> bool:
        """
        :return: True if thread is cancelled, False otherwise.
        """

        return self._cancel_event.is_set()

    def cancel(self):
        """Cancels the thread."""

        self._cancel_event.set()



from copy import copy
from typing import List


class RunVersion:
    """
    The current version of a run, using the identification scheme of the
    Shared Versioned Match Buffer, as proposed by Agrawal et al. (2008)
    in their paper "Efficient Pattern Matching over Event Streams".

    :param parent_version: The parent version. If provided, the new version
                           will copy the parent version's levels.
                           Defaults to None.
    :type parent_version: RunVersion
    """

    _DELIMITER = "."

    def __init__(self, parent_version: 'RunVersion' = None) -> None:
        super().__init__()

        self._levels = [] if parent_version is None \
            else copy(parent_version._levels)

    @staticmethod
    def list_to_version(version_list: List[str]) -> 'RunVersion':
        """
        Converts a list of strings into a RunVersion instance, where each
        string of the list becomes a level in the RunVersion instance, in list
        order.

        :param version_list: A list of strings.
        :type version_list: List[str]

        :return: A new RunVersion instance.
        """

        run = RunVersion()
        run._levels = version_list
        return run

    @staticmethod
    def list_to_version_str(str_list: List[str]) -> str:
        """
        Converts a list of strings into a valid version string by joining
        each string in the list by the delimeter ".".

        :param str_list: A list of strings.
        :type str_list: List[str]

        :return: A valid version string.
        """

        return RunVersion._DELIMITER.join(str_list)

    @staticmethod
    def str_to_version(version_str: str) -> 'RunVersion':
        """
        Converts a version string into a RunVersion instance by splitting the
        string by the delimiter "." and making each split a level in the new
        instance.

        :param version_str: The version string.
        :type version_str: str

        :return: A new RunVersion instance.
        """

        run = RunVersion()
        run._levels = version_str.split(RunVersion._DELIMITER)
        return run

    def add_level(self, level: str) -> None:
        """
        Adds a new level to the version.

        :param level: The level name.
        :type level: str
        """

        self._levels.append([level])

    def increment_level(self, increment: str) -> None:
        """
        Increments the current level.

        :param increment: The increment name.
        :type increment: str

        :raises RuntimeError: Attempting to increment when there are no levels.
        """

        if len(self._levels) == 0:
            raise RuntimeError("Cannot increment. No levels found.")

        self._levels[-1].append(increment)

    def size(self) -> int:
        """
        :return: The number of levels.
        """

        return len(self._levels)

    def size_level(self, index: int = -1):
        """
        :param index: The index of the level, defaults to the latest level.
        :type index: int, optional

        :return: The number of increments in at a given level.
        """

        return len(self._levels[index])

    def remove_all_levels(self) -> None:
        """
        Removes all levels from the version.
        """

        self._levels = []

    def get_version_as_list(self) -> List[str]:
        """
        :return: The latest version as a list of strings, where each string
                 is the latest increment of the given level.
        """

        return [level[-1] for level in self._levels]

    def get_version_as_str(self) -> str:
        """
        :return: The latest version as a string, where each level is the latest
                 increment of the given level, separated with the delimeter
                 ".", for example, "a.b.c".
                 If there are no levels, it returns an empty string.
        """

        if len(self._levels) == 0:
            return ""

        return RunVersion._DELIMITER.join(
            [level[-1] for level in self._levels])

    def get_previous_version_as_list(self,
                                     decrease_level: int,
                                     decrease_incr: int) -> List[str]:
        """
        Gets a previous version of the run version as a list of strings.

        :param decrease_level: How many levels to decrease by.
                               For example, if there are 5 levels, a
                               decrease_level of 2 will go back to level 3.
        :type decrease_level: int

        :param decrease_incr: How many increments of a level to decrease by.
        :type decrease_incr: int

        :raises RuntimeError: Decreasing beyond the number of levels.
        :raises RuntimeError: Decreasing beyond the number of increments in
                              the chosen level.

        :return: The previous version as a list.
        """

        index_level = len(self._levels) - decrease_level - 1

        if not (0 <= index_level < len(self._levels)):
            raise RuntimeError("Level decrease is greater than the number of "
                               "levels. Version has {} levels, decrease is {}."
                               .format(len(self._levels), decrease_level))

        index_incr = len(self._levels[index_level]) - decrease_incr - 1

        if not (0 <= index_incr < len(self._levels[index_level])):
            raise RuntimeError("Increment decrease is greater than the number "
                               "of increments in level {}. "
                               "Level has {} increments, decrease is {}."
                               .format(index_level + 1,
                                       len(self._levels[index_level]),
                                       decrease_incr))

        version = [level[-1] for level in self._levels[:index_level]]
        version.append(self._levels[index_level][index_incr])

        return version

    def get_previous_version_as_str(self,
                                    decrease_level: int,
                                    decrease_incr: int,
                                    default=None) -> str:
        """
        Gets a previous version of the run version as a string, delimited by
        ".".

        :param decrease_level: How many levels to decrease by.
                               For example, if there are 5 levels, a
                               decrease_level of 2 will go back to level 3.
        :type decrease_level: int

        :param decrease_incr: How many increments of a level to decrease by.
        :type decrease_incr: int

        :param default: The default value to return if a version string fails
                        to be generated, defaults to None.
        :type default: any, optional

        :return: The previous version as a string delimited by ".".
                 If there are no levels, it returns an empty string.
        """

        if len(self._levels) == 0:
            return ""

        try:
            return RunVersion._DELIMITER.join(
                self.get_previous_version_as_list(decrease_level,
                                                  decrease_incr))
        except RuntimeError:
            return default



from abc import ABC

from bobocep.rules.actions.bobo_action import BoboAction


class MultiAction(BoboAction, ABC):
    """An abstract action that enables the execution of multiple actions
    within a single BoboAction instance.

    :param name: The action name, defaults to an empty string.
    :type name: str, optional
    """

    def __init__(self, name: str = None) -> None:
        super().__init__(name=name)



from typing import Dict, List

from bobocep.rules.predicates.bobo_predicate import BoboPredicate
from bobocep.rules.states.bobo_state import BoboState
from bobocep.rules.transitions.bobo_transition import BoboTransition


class BoboNFA:
    """A :code:`bobocep` nondeterministic finite automaton.

    :param name: The automaton name.
    :type name: str

    :param states: The states of the automaton.
    :type states: Dict[str, BoboState]

    :param transitions: The transitions between states.
    :type transitions: Dict[str, BoboTransition]

    :param start_state_name: The name of the start state.
    :type start_state_name: str

    :param final_state_name: The name of the final state.
    :type final_state_name: str

    :param preconditions: Predicates checked before any state's predicate
                          whereby, if *any* precondition returns False,
                          the automaton halts.
    :type preconditions: List[BoboPredicate]

    :param haltconditions: Predicates checked before any state's predicate
                           whereby, if *any* haltcondition returns True,
                           the automaton halts.
    :type haltconditions: List[BoboPredicate]
    """

    def __init__(self,
                 name: str,
                 states: Dict[str, BoboState],
                 transitions: Dict[str, BoboTransition],
                 start_state_name: str,
                 final_state_name: str,
                 preconditions: List[BoboPredicate],
                 haltconditions: List[BoboPredicate]) -> None:
        super().__init__()

        self.name = name
        self.states = states
        self.transitions = transitions
        self.start_state = states[start_state_name]
        self.final_state = states[final_state_name]
        self.preconditions = preconditions
        self.haltconditions = haltconditions

        self.start_is_final = start_state_name == final_state_name



from abc import ABC, abstractmethod

from bobocep.rules.events.bobo_event import BoboEvent
from bobocep.rules.events.composite_event import CompositeEvent


class INFAHandlerSubscriber(ABC):

    @abstractmethod
    def on_handler_transition(self,
                              nfa_name: str,
                              run_id: str,
                              state_name_from: str,
                              state_name_to: str,
                              event: BoboEvent):
        """
        Triggers a response when a state transition occurs in a run.

        :param nfa_name: The NFA name.
        :type nfa_name: str

        :param run_id: The run ID.
        :type run_id: str

        :param state_name_from: The original state before the transition.
        :type state_name_from: str

        :param state_name_to: The new state after the transition.
        :type state_name_to: str

        :param event: The event that caused the state transition.
        :type event: BoboEvent
        """

    @abstractmethod
    def on_handler_clone(self,
                         nfa_name: str,
                         run_id: str,
                         state_name: str,
                         event: BoboEvent):
        """Triggers a response when a clone occurs in a run. Newly created runs
        are considered to be cloned but without a parent run.

        :param nfa_name: The NFA name.
        :type nfa_name: str

        :param run_id: The cloned run ID.
        :type run_id: str

        :param state_name: The state of the cloned run.
        :type state_name: str

        :param event: The event that caused the clone.
        :type event: BoboEvent
        """

    @abstractmethod
    def on_handler_final(self,
                         nfa_name: str,
                         run_id: str,
                         event: CompositeEvent):
        """Triggers a response when a run reaches its final state.

        :param nfa_name: The NFA name.
        :type nfa_name: str

        :param run_id: The run ID.
        :type run_id: str

        :param event: The complex event.
        :type event: CompositeEvent
        """

    @abstractmethod
    def on_handler_halt(self,
                        nfa_name: str,
                        run_id: str):
        """Triggers a response when a run halts. This occurs if a run reaches
        its final state, or if a run halts prematurely.

        :param nfa_name: The NFA name.
        :type nfa_name: str

        :param run_id: The run ID.
        :type run_id: str
        """



TRANSITION = "transition"
CLONE = "clone"
HALT = "halt"
FINAL = "final"
ACTION = "action"

SYNC_REQ = "sync_req"
SYNC_RES = "sync_res"

NFA_NAME = "nfa_name"
RUN_ID = "run_id"
STATE_FROM = "state_from"
STATE_TO = "state_to"
EVENT = "event"
HISTORY = "history"
HANDLERS = "handlers"



from abc import ABC, abstractmethod
from queue import Queue

from bobocep.decider.decider_subscriber import IDeciderSubscriber
from bobocep.producer.producer_subscriber import IProducerSubscriber
from bobocep.rules.actions.action_subscriber import IActionSubscriber
from bobocep.rules.events.action_event import ActionEvent
from bobocep.rules.events.composite_event import CompositeEvent
from bobocep.setup.task.bobo_task import BoboTask


class BoboProducer(BoboTask,
                   IDeciderSubscriber,
                   IActionSubscriber,
                   ABC):
    """A :code:`bobocep` event producer that creates CompositeEvent instances
    when a run reaches its final state.

    :param max_queue_size: The maximum data queue size,
                           defaults to 0 (infinite).
    :type max_queue_size: int, optional

    :param active: Whether task should start in an active state,
                   defaults to True.
    :type active: bool, optional
    """

    def __init__(self, max_queue_size: int = 0, active: bool = True) -> None:
        super().__init__(active=active)

        self._event_queue = Queue(maxsize=max_queue_size)
        self._subs = {}

    def _loop(self) -> None:
        if not self._event_queue.empty():
            event = self._event_queue.get_nowait()

            if event is not None:
                if self._handle_producer_event(event):
                    for subscriber in self._subs[event.name]:
                        subscriber.on_accepted_producer_event(event)
                else:
                    for subscriber in self._subs[event.name]:
                        subscriber.on_rejected_producer_event(event)

    @abstractmethod
    def _handle_producer_event(self, event: CompositeEvent) -> bool:
        """
        Enables the producer to perform some action with the CompositeEvent
        instance before it is sent to the subscribers of the producer.
        If the method returns True, the CompositeEvent instance will be
        forwarded to subscribers.

        :param event: The event.
        :type event: CompositeEvent

        :return: True if event was successfully handled, False otherwise.
        """

    def on_decider_complex_event(self, event: CompositeEvent):
        if not self._is_cancelled:
            self._event_queue.put_nowait(event)

    def on_action_attempt(self, event: ActionEvent):
        if not self._is_cancelled:
            if isinstance(event.for_event, CompositeEvent):
                for subscriber in self._subs[event.for_event.name]:
                    subscriber.on_producer_action(event)

    def subscribe(self,
                  event_name: str,
                  subscriber: IProducerSubscriber) -> None:
        """
        :param event_name: Subscribe to CompositeEvent instances which have
                           this name.
        :type event_name: str

        :param subscriber: Subscriber to events from Producer.
        :type subscriber: IProducerSubscriber
        """

        with self._lock:
            if not self._is_cancelled:
                if event_name not in self._subs:
                    self._subs[event_name] = []

                if subscriber not in self._subs[event_name]:
                    self._subs[event_name].append(subscriber)

    def unsubscribe(self,
                    event_name: str,
                    unsubscriber: IProducerSubscriber) -> None:
        """
        :param event_name: Unsubscribe to CompositeEvent instances which have
                           this name.
        :type event_name: str

        :param unsubscriber: Unsubscriber to events from Producer.
        :type unsubscriber: IProducerSubscriber
        """

        with self._lock:
            if event_name in self._subs:
                if unsubscriber in self._subs[event_name]:
                    self._subs[event_name].remove(unsubscriber)

    def _setup(self) -> None:
        """"""

    def _cancel(self) -> None:
        """"""



from abc import ABC, abstractmethod

from bobocep.rules.events.composite_event import CompositeEvent


class IDeciderSubscriber(ABC):
    """An interface to subscribe to Decider events."""

    @abstractmethod
    def on_decider_complex_event(self, event: CompositeEvent):
        """
        When a Decider identifies a complex event in the event stream.

        :param event: The complex event.
        :type event: CompositeEvent
        """



from threading import RLock
from time import time_ns
from typing import List
from uuid import uuid4

from bobocep.decider.bobo_decider import BoboDecider
from bobocep.decider.buffers.shared_versioned_match_buffer import \
    SharedVersionedMatchBuffer
from bobocep.decider.decider_subscriber import IDeciderSubscriber
from bobocep.decider.handlers.bobo_nfa_handler import BoboNFAHandler
from bobocep.forwarder.action_forwarder import ActionForwarder
from bobocep.forwarder.forwarder_subscriber import IForwarderSubscriber
from bobocep.producer.action_producer import ActionProducer
from bobocep.producer.producer_subscriber import IProducerSubscriber
from bobocep.receiver.bobo_receiver import BoboReceiver
from bobocep.receiver.formatters.primitive_event_formatter import \
    PrimitiveEventFormatter
from bobocep.receiver.generators.bobo_null_data_generator import \
    BoboNullDataGenerator
from bobocep.receiver.generators.data.bobo_null_data import BoboNullData
from bobocep.receiver.receiver_subscriber import IReceiverSubscriber
from bobocep.receiver.validators.any_validator import AnyValidator
from bobocep.receiver.validators.bobo_validator import \
    BoboValidator
from bobocep.rules.actions.bobo_action import BoboAction
from bobocep.rules.actions.no_action import NoAction
from bobocep.rules.bobo_rule_builder import BoboRuleBuilder
from bobocep.setup.bobo_complex_event import \
    BoboComplexEvent
from bobocep.setup.distributed.bobo_dist_manager import BoboDistManager
from bobocep.setup.distributed.outgoing.dist_outgoing_subscriber import \
    IDistOutgoingSubscriber
from bobocep.setup.task.bobo_task_thread import BoboTaskThread


class BoboSetup(IDistOutgoingSubscriber):
    """A class to set up a working :code:`bobocep` process.

    :param delay: The delay, in seconds, for internal threads to wait before
                  repeating its processing task, defaults to 0.5.
    :type delay: float, optional

    :param recursive: Whether CompositeEvent instances should be put back
                      into the system on generation, defaults to True.
    :type recursive: bool, optional

    :param max_queue_size: The maximum data queue size,
                           defaults to 0 (infinite).
    :type max_queue_size: int, optional

    :param max_recent: The maximum number of recent CompositeEvent instances
                        generated by this handler to pass to runs.
                        Minimum of 1, defaults to 1.
    :type max_recent: int, optional
    """

    def __init__(self,
                 delay: float = 0.5,
                 recursive: bool = True,
                 max_queue_size: int = 0,
                 max_recent: int = 1) -> None:
        super().__init__()

        self._delay = delay
        self._recursive = recursive
        self._max_queue_size = max_queue_size
        self._max_recent = max(1, max_recent)

        self._lock = RLock()
        self._running = False
        self._cancelled = False
        self._configured = False

        self._validator = None
        self._event_defs = []
        self._action_producer = None
        self._action_forwarder = None

        self._req_null_data = False
        self._null_data_generator = None
        self._null_data_delay = 0
        self._null_data_obj = None

        self._distributed = False
        self._manager = None
        self._exchange_name = None
        self._user_name = None
        self._host_name = None
        self._user_id = None
        self._synced = False

        self._receiver = None
        self._decider = None
        self._producer = None
        self._forwarder = None

        self._receiver_thread = None
        self._null_event_thread = None
        self._decider_thread = None
        self._producer_thread = None
        self._forwarder_thread = None

        self._subs_receiver = []
        self._subs_decider = {}
        self._subs_producer = {}
        self._subs_forwarder = []

    def is_ready(self) -> bool:
        """
        :return: True if active and synced, False otherwise.
        """

        with self._lock:
            if self.is_active():
                if self._distributed:
                    return self._synced
                else:
                    return True

    def is_active(self) -> bool:
        """
        :return: True if running and not cancelled, False otherwise.
        """

        with self._lock:
            return self._running and not self._cancelled

    def is_inactive(self) -> bool:
        """
        :return: True if not running and not cancelled, False otherwise.
        """

        with self._lock:
            return not self._running and not self._cancelled

    def is_cancelled(self) -> bool:
        """
        :return: True if cancelled, False otherwise.
        """

        with self._lock:
            return self._cancelled

    def is_configured(self) -> bool:
        """
        :return: True if configured, False otherwise.
        """

        with self._lock:
            return self._configured

    def get_receiver(self) -> BoboReceiver:
        """
        :raises RuntimeError: Attempting access before configuration.

        :return: The Receiver.
        """

        with self._lock:
            if not self._configured:
                raise RuntimeError("Setup must be configured first.")

            return self._receiver

    def get_decider(self) -> BoboDecider:
        """
        :raises RuntimeError: Attempting access before configuration.

        :return: The Receiver.
        """

        with self._lock:
            if not self._configured:
                raise RuntimeError("Setup must be configured first.")

            return self._decider

    def get_producer(self) -> ActionProducer:
        """
        :raises RuntimeError: Attempting access before configuration.

        :return: The Receiver.
        """

        with self._lock:
            if not self._configured:
                raise RuntimeError("Setup must be configured first.")

            return self._producer

    def get_forwarder(self) -> ActionForwarder:
        """
        :raises RuntimeError: Attempting access before configuration.

        :return: The Forwarder.
        """

        with self._lock:
            if not self._configured:
                raise RuntimeError("Setup must be configured first.")

            return self._forwarder

    def get_distributed(self) -> BoboDistManager:
        """
        :raises RuntimeError: Attempting access before configuration.
        :raises RuntimeError: Distributed was not configured.

        :return: The distrubted manager.
        """

        with self._lock:
            if not self._configured:
                raise RuntimeError("Setup must be configured first.")
            elif not self._distributed:
                raise RuntimeError("Distributed was not configured.")

            return self._manager

    def get_null_data_generator(self) -> BoboNullDataGenerator:
        """
        :raises RuntimeError: Attempting access before configuration.

        :return: The null data generator, or None if one was not configured
         during setup.
        """

        with self._lock:
            if not self._configured:
                raise RuntimeError("Setup must be configured first.")

            return self._null_data_generator

    def get_complex_events(self) -> List[BoboComplexEvent]:
        """
        :return: A list of complex event definitions currently in setup.
        """

        with self._lock:
            if self.is_inactive():
                return self._event_defs[:]

    def add_complex_event(self, event_def: BoboComplexEvent) -> None:
        """
        Adds a complex event definition to the setup.

        :param event_def: The complex event definition.
        :type event_def: BoboComplexEvent
        """

        with self._lock:
            if self.is_inactive():
                self._event_defs.append(event_def)

    def config_receiver(self, validator: BoboValidator) -> None:
        """
        Configure the Receiver.

        :param validator: The validator to use in the Receiver.
        :type validator: BoboValidator
        """

        with self._lock:
            if self.is_inactive():
                self._validator = validator

    def config_producer(self, action: BoboAction) -> None:
        """
        Configure the Producer.

        :param action: The action to execute in the Producer on CompositeEvent
                       generation. If this action returns True, the
                       CompositeEvent is passed to the Forwarder. Otherwise,
                       it is dropped.
        :type action: BoboAction
        """

        with self._lock:
            if self.is_inactive():
                self._action_producer = action

    def config_forwarder(self, action: BoboAction) -> None:
        """
        Configure the Forwarder.

        :param action: The action to execute in the Forwarder. If this action
                       returns True, the CompositeEvent is passed to the
                       Forwarder's subscribers. Otherwise, it is dropped.
        :type action: BoboAction
        """

        with self._lock:
            if self.is_inactive():
                self._action_forwarder = action

    def config_distributed(self,
                           exchange_name: str,
                           user_name: str,
                           host_name: str) -> None:
        """
        Configure the connection to the external message broker.

        :param exchange_name: The exchange name.
        :type exchange_name: str

        :param user_name: The user name.
        :type user_name: str

        :param host_name: The host name.
        :type host_name: str
        """

        with self._lock:
            if self.is_inactive():
                self._distributed = True
                self._exchange_name = exchange_name
                self._user_name = user_name
                self._host_name = host_name

    def config_null_data(self,
                         delay_sec: float,
                         null_data: BoboNullData) -> None:
        """
        Configure static data to be input periodically into the Receiver.

        :param delay_sec: The rate at which to send data. Minimum of 0.1
                          seconds.
        :type delay_sec: float

        :param null_data: The data to send. Ensure that it is able to pass the
                          Receiver's validation criteria. Otherwise, the null
                          data will not enter the system.
        :type null_data: BoboNullData
        """

        with self._lock:
            if self.is_inactive():
                self._req_null_data = True
                self._null_data_delay = max(0.1, delay_sec)
                self._null_data_obj = null_data

    def on_sync(self) -> None:
        with self._lock:
            if self._distributed:
                self._activate_tasks()
            self._synced = True

    def start(self) -> None:
        """Start the setup.

        :raises RuntimeError: Running setup when it is already active.
        :raises RuntimeError: No complex event definitions found.
        """

        with self._lock:
            if self.is_active():
                raise RuntimeError("Setup is already active.")

            if not self._configured:
                self.configure()
            self._start_threads()
            self._running = True

            # tasks active by default if not distributed, sync immediately
            if not self._distributed:
                self._synced = True

    def configure(self) -> None:
        with self._lock:
            if self._cancelled:
                raise RuntimeError("Setup has been cancelled.")

            if self.is_active():
                raise RuntimeError("Setup is already active.")

            if self._configured:
                raise RuntimeError("Setup is already configured.")

            if len(self._event_defs) == 0:
                raise RuntimeError(
                    "No complex event definitions found. "
                    "At least one must be provided.")

            self._config_receiver()
            self._config_decider()
            self._config_producer()
            self._config_forwarder()

            self._config_distributed()
            self._config_definitions()
            self._config_extra_subscriptions()

            self._configured = True

    def cancel(self) -> None:
        """Cancel the setup."""

        with self._lock:
            if not self._cancelled:
                self._cancelled = True
                if self._running:
                    self._running = False
                    self._deactivate_tasks()
                    self._stop_threads()

    def subscribe_receiver(self, subscriber: IReceiverSubscriber) -> None:
        with self._lock:
            if self.is_inactive():
                if subscriber not in self._subs_receiver:
                    self._subs_receiver.append(subscriber)

    def subscribe_decider(self,
                          nfa_name: str,
                          subscriber: IDeciderSubscriber) -> None:
        with self._lock:
            if self.is_inactive():
                if nfa_name not in self._subs_decider:
                    self._subs_decider[nfa_name] = []

                if subscriber not in self._subs_decider[nfa_name]:
                    self._subs_decider[nfa_name].append(subscriber)

    def subscribe_producer(self,
                           nfa_name: str,
                           subscriber: IProducerSubscriber) -> None:
        with self._lock:
            if self.is_inactive():
                if nfa_name not in self._subs_producer:
                    self._subs_producer[nfa_name] = []

                if subscriber not in self._subs_producer[nfa_name]:
                    self._subs_producer[nfa_name].append(subscriber)

    def subscribe_forwarder(self, subscriber: IForwarderSubscriber) -> None:
        with self._lock:
            if self.is_inactive() and subscriber not in self._subs_forwarder:
                self._subs_forwarder.append(subscriber)

    def _config_receiver(self) -> None:
        if self._validator is None:
            self._validator = AnyValidator()

        self._receiver = BoboReceiver(
            validator=self._validator,
            formatter=PrimitiveEventFormatter(),
            max_queue_size=self._max_queue_size,
            active=(not self._distributed))

        self._receiver_thread = BoboTaskThread(
            task=self._receiver,
            delay=self._delay)

        if self._req_null_data:
            self._null_data_generator = BoboNullDataGenerator(
                receiver=self._receiver,
                null_data=self._null_data_obj,
                active=(not self._distributed))

            self._null_event_thread = BoboTaskThread(
                task=self._null_data_generator,
                delay=self._null_data_delay
            )

    def _config_decider(self) -> None:
        self._decider = BoboDecider(
            recursive=self._recursive,
            max_queue_size=self._max_queue_size,
            active=(not self._distributed))

        self._decider_thread = BoboTaskThread(
            task=self._decider,
            delay=self._delay)

        # Receiver -> Decider
        self._receiver.subscribe(self._decider)

    def _config_producer(self) -> None:
        if self._action_producer is None:
            self._action_producer = NoAction()

        self._producer = ActionProducer(
            action=self._action_producer,
            max_queue_size=self._max_queue_size,
            active=(not self._distributed))

        self._producer_thread = BoboTaskThread(
            task=self._producer,
            delay=self._delay)

    def _config_forwarder(self) -> None:
        if self._action_forwarder is None:
            self._action_forwarder = NoAction()

        self._forwarder = ActionForwarder(
            action=self._action_forwarder,
            max_queue_size=self._max_queue_size,
            active=(not self._distributed))

        self._forwarder_thread = BoboTaskThread(
            task=self._forwarder,
            delay=self._delay)

    def _config_distributed(self) -> None:
        if self._distributed:
            self._user_id = "{}-{}-{}".format(
                self._user_name,
                uuid4(),
                time_ns())

            self._manager = BoboDistManager(
                decider=self._decider,
                exchange_name=self._exchange_name,
                user_id=self._user_id,
                host_name=self._host_name,
                delay=self._delay)

            # setup will activate tasks on sync
            self._manager.outgoing.subscribe(self)

    def _config_definitions(self) -> None:
        for event_def in self._event_defs:
            handler = BoboNFAHandler(
                nfa=BoboRuleBuilder.nfa(
                    name_nfa=event_def.name,
                    pattern=event_def.pattern),
                buffer=SharedVersionedMatchBuffer(),
                max_recent=self._max_recent)

            self._decider.add_nfa_handler(handler)

            # Decider -> Producer
            self._decider.subscribe(event_def.name, self._producer)

            if self._recursive:
                # Producer -> Decider
                self._producer.subscribe(event_def.name, self._decider)

            # Producer -> Forwarder
            self._producer.subscribe(event_def.name, self._forwarder)

            if event_def.action is not None:
                # Producer -> Action
                self._producer.subscribe(event_def.name, event_def.action)

                # Action -> Producer
                event_def.action.subscribe(self._producer)

            if self._distributed:
                # Handler -> Outgoing
                handler.subscribe(self._manager.outgoing)

                if event_def.action is not None:
                    # Action -> Outgoing
                    event_def.action.subscribe(self._manager.outgoing)

                # Incoming -> Handler
                self._manager.incoming.subscribe(handler)

    def _config_extra_subscriptions(self):
        # receiver
        for sub in self._subs_receiver:
            self._receiver.subscribe(sub)

        # decider
        for nfa_name, subs in self._subs_decider.items():
            for sub in subs:
                self._decider.subscribe(nfa_name, sub)

        # producer
        for nfa_name, subs in self._subs_producer.items():
            for sub in subs:
                self._producer.subscribe(nfa_name, sub)

        # forwarder
        for sub in self._subs_forwarder:
            self._forwarder.subscribe(sub)

    def _start_threads(self) -> None:
        self._receiver_thread.start()
        self._decider_thread.start()
        self._producer_thread.start()
        self._forwarder_thread.start()

        if self._null_event_thread is not None:
            self._null_event_thread.start()

        if self._manager is not None:
            self._manager.start()

    def _stop_threads(self) -> None:
        self._receiver_thread.cancel()
        self._decider_thread.cancel()
        self._producer_thread.cancel()
        self._forwarder_thread.cancel()

        if self._null_event_thread is not None:
            self._null_event_thread.cancel()

        if self._manager is not None:
            self._manager.cancel()

    def _activate_tasks(self) -> None:
        self._forwarder.activate()
        self._producer.activate()
        self._decider.activate()
        self._receiver.activate()

        if self._null_data_generator is not None:
            self._null_data_generator.activate()

    def _deactivate_tasks(self) -> None:
        if self._null_data_generator is not None:
            self._null_data_generator.deactivate()

        self._receiver.deactivate()
        self._decider.deactivate()
        self._producer.deactivate()
        self._forwarder.deactivate()



from bobocep.receiver.validators.bobo_validator import \
    BoboValidator


class StrDictValidator(BoboValidator):
    """A validator that checks whether data is a dict, where both the keys and
    values are of type str.

    :param min_length: The minimum length for strings, defaults to 0.
    :type min_length: int, optional
    """

    def __init__(self, min_length: int = 0) -> None:
        super().__init__()

        self._min_length = max(0, min_length)

    def validate(self, data) -> bool:
        if type(data) is not dict:
            return False

        for key, value in data.items():
            if type(key) is not str or type(value) is not str:
                return False

            if len(key) < self._min_length or len(value) < self._min_length:
                return False

        return True



from copy import copy
from typing import List


class BoboTransition:
    """A :code:`bobocep` state transition.

    :param state_names: The names of states to which a transition can be made,
                        defaults to an empty list.
    :type state_names: List[str], optional

    :param strict: Whether strict contiguity is expected for the transition,
                   defaults to False.
    :type strict: bool, optional

    :raises RuntimeError: Using strict contiguity when more than one state is
                          provided.
    """

    STATE_NAMES = "state_names"
    STRICT = "strict"

    def __init__(self,
                 state_names: List[str] = None,
                 strict: bool = False) -> None:
        super().__init__()

        if len(state_names) > 1 and strict:
            raise RuntimeError("Must not use strict contiguity with more than "
                               "one state, found {} states."
                               .format(len(state_names)))

        self.state_names = [] if state_names is None else state_names
        self.is_strict = strict
        self.is_deterministic = len(state_names) == 1

    def to_dict(self) -> dict:
        """
        :return: A dict representation of the object.
        """

        return {
            self.STATE_NAMES: copy(self.state_names),
            self.STRICT: self.is_strict
        }



from abc import ABC

from bobocep.rules.predicates.bobo_predicate import BoboPredicate


class BoboPredicateWindow(BoboPredicate, ABC):
    """A predicate that evaluates using a time window."""

    def __init__(self) -> None:
        super().__init__()



from bobocep.receiver.generators.data.bobo_null_data import BoboNullData


class BoboNullDataStatic(BoboNullData):
    """
    Null data class that always returns some static data.
    """

    def __init__(self, data) -> None:
        super().__init__()

        self._data = data

    def get_null_data(self):
        return self._data



from abc import ABC, abstractmethod

from bobocep.rules.events.action_event import ActionEvent


class IActionSubscriber(ABC):
    """An interface to subscribe to Action events."""

    @abstractmethod
    def on_action_attempt(self, event: ActionEvent):
        """
        When an action has been attempted.

        :param event: The action event.
        :type event: ActionEvent
        """



from abc import ABC, abstractmethod
from typing import List

from bobocep.rules.events.bobo_event import BoboEvent
from bobocep.rules.events.histories.bobo_history import BoboHistory
from bobocep.rules.predicates.windows.bobo_predicate_window import \
    BoboPredicateWindow


class WindowSliding(BoboPredicateWindow, ABC):
    """A predicate that evaluates using a sliding time window.

    :param interval_ns: The sliding window time interval, in nanoseconds.
    :type interval_ns: int
    """

    def __init__(self, interval_ns: int) -> None:
        super().__init__()

        self._interval_ns = interval_ns

    @abstractmethod
    def get_previous_event(self, history: BoboHistory) -> BoboEvent:
        """Gets a previous event in the event history, for use in calculating
        the time interval between two events.

        :param history: The event history.
        :type history: BoboHistory

        :return: A historical event.
        """

    def evaluate(self,
                 event: BoboEvent,
                 history: BoboHistory,
                 recent: List[BoboEvent]) -> bool:
        event_history = self.get_previous_event(history)
        return (event.timestamp -
                event_history.timestamp) <= self._interval_ns



from typing import List

from bobocep.rules.predicates.bobo_predicate import BoboPredicate


class BoboPatternLayer:
    """A BoboPattern layer, describing the state(s) of a single layer of an
    automaton.

    :param label: The label with which the state(s) will be associated.
    :type label: str

    :param predicates: The predicate(s) that the state(s) will use for
                       evaluation.
    :type predicates: List[BoboPredicate]

    :param times: How many copies of the state(s) to have in sequence,
                  default to 1.
    :type times: int

    :param loop: Whether the state(s) is/are self-looping.
    :type loop: bool

    :param strict: Whether the state(s) has/have strict contiguity.
    :type strict: bool

    :param negated: Whether the state(s) is/are negated.
    :type negated: bool

    :param optional: Whether the state(s) is/are optional.
    :type optional: bool
    """

    def __init__(self,
                 label: str,
                 predicates: List[BoboPredicate],
                 times: int,
                 loop: bool,
                 strict: bool,
                 negated: bool,
                 optional: bool) -> None:
        super().__init__()

        self.label = label
        self.predicates = predicates
        self.times = times
        self.loop = loop
        self.strict = strict
        self.negated = negated
        self.optional = optional



from abc import ABC, abstractmethod
from threading import RLock


class BoboTask(ABC):
    """A task that is scheduled to repeat indefinitely until cancelled.

    :param active: Whether task should start in an active state,
                   defaults to True.
    :type active: bool, optional
    """

    def __init__(self, active: bool = True) -> None:
        super().__init__()

        self._is_active = active
        self._is_cancelled = False
        self._is_setup = False
        self._lock = RLock()

    @abstractmethod
    def _setup(self) -> None:
        """"""

    @abstractmethod
    def _loop(self) -> None:
        """"""

    @abstractmethod
    def _cancel(self) -> None:
        """"""

    def activate(self) -> None:
        """Activates task."""

        with self._lock:
            self._is_active = True

    def deactivate(self) -> None:
        """Deactivates task."""

        with self._lock:
            self._is_active = False

    def is_active(self) -> bool:
        """
        :return: True if the task is current active, False otherwise.
        """

        with self._lock:
            return self._is_active

    def cancel(self) -> None:
        """Cancel the task."""

        with self._lock:
            if not self._is_cancelled:
                self._cancel()
                self._is_cancelled = True

    def is_cancelled(self) -> bool:
        """
        :return: True if the task has been cancelled, False otherwise.
        """

        with self._lock:
            return self._is_cancelled

    def setup(self) -> None:
        """Perform initial setup. It can only be performed once, and must be
        performed before loop functionality can occur.

        :raises RuntimeError: Task has already been cancelled.
        :raises RuntimeError: Setup has already been performed.
        :raises RuntimeError: Exception raised during setup.
        """

        with self._lock:
            if self._is_cancelled:
                raise RuntimeError("Task has already been cancelled.")

            if self._is_setup:
                raise RuntimeError("Setup has already been performed.")

            try:
                self._setup()
            except Exception as e:
                raise RuntimeError("Exception raised during setup: {}."
                                   .format(e))
            self._is_setup = True

    def is_setup(self) -> bool:
        """
        :return: True if task has successfully set up, False otherwise.
        """

        with self._lock:
            return self._is_setup

    def loop(self) -> None:
        """Perform primary task loop. The task must not be cancelled, and
        setup must have already occurred.

        :raises RuntimeError: Task has already been cancelled.
        :raises RuntimeError: Setup must be performed before using loop.
        :raises RuntimeError: Exception raised during loop.
        """

        with self._lock:
            if self._is_cancelled:
                raise RuntimeError("Task has already been cancelled.")

            if not self._is_setup:
                raise RuntimeError(
                    "Setup must be performed before using loop.")

            if self._is_active:
                try:
                    self._loop()
                except Exception as e:
                    raise RuntimeError("Exception raised during loop: {}."
                                       .format(e))
